#!/usr/bin/env python3
# nuitka-project: --follow-import-to=py2stl
# nuitka-project: --follow-import-to=wire3d

from __future__ import division
from __future__ import print_function

import sys
import os.path

import pygame
import py2stl
import wire3d

from math import radians

import subprocess
import shutil

SCREENW = 800
SCREENH = 600
CENTERX = SCREENW // 2
CENTERY = SCREENH // 2
FSIZE = SCREENH // 20

ROTSTEP = 15

DARKSLATEGRAY = (47, 79, 79)
SEASHELL = (255, 245, 238)
GHOSTWHITE = (248, 248, 255)

pygame.init()

screen = pygame.display.set_mode((SCREENW, SCREENH), pygame.RESIZABLE)
pygame.display.set_caption("WireView 3D")
pygame.event.set_blocked(pygame.MOUSEMOTION)

font_name = pygame.font.get_default_font()
hud_font = pygame.font.SysFont(font_name, FSIZE)

def renderer(x1, y1, x2, y2):
	x1 += CENTERX
	y1 = CENTERY - y1
	x2 += CENTERX
	y2 = CENTERY - y2
	pygame.draw.line(screen, GHOSTWHITE, (x1, y1), (x2, y2))

def drawstr(msg, x, y):
	text = hud_font.render(msg, False, SEASHELL)
	text_pos = text.get_rect(topleft = (x, y))
	screen.blit(text, text_pos)

def count_msg(model):
	msg = "Loaded model with {0} vertices and {1} faces."
	return msg.format(model.vertex_count(), model.face_count())

camera = wire3d.Camera(0, 0, 0, SCREENH)
model = wire3d.Group()
bb_c = wire3d.Vertex(0, 0, 0)

angle_x = 0
angle_z = 0
zoom_step = 5

error_msg = "No model loaded."

def load_file(name):
	global model, error_msg
	with open(name) as f:
		if name.lower().endswith(".stl"):
			parser = wire3d.STLparser(f)
			s = wire3d.Shape()
			while parser.parse_solid(s):
				model += s
				s = wire3d.Shape()
		elif name.lower().endswith(".obj"):
			model += wire3d.load_obj(f)
		else:
			raise RuntimeError("Unknown file type")
	fn = os.path.basename(name)
	pygame.display.set_caption(fn + " | WireView 3D")
	error_msg = count_msg(model)

def setup_view():
	global zoom_step
	m_min = model.bbox_min()
	m_max = model.bbox_max()
	bb_c = (m_min + m_max) / 2
	size = m_max - m_min
	camera.x = bb_c.x
	camera.y = bb_c.z
	camera.z = bb_c.y - max(size.x, size.z) * 1.5
	if camera.z >= m_min.y:
		camera.z = bb_c.y - size.y
	zoom_step = abs(camera.z - bb_c.y) / 5

zenity_cmd = ["zenity", "--file-selection",
	"--title", "Open 3D model",
	"--file-filter", "All files | *",
	"--file-filter", "STL files | *.stl",
	"--file-filter", "OBJ files | *.obj"]

def open_with_zenity():
	global model, error_msg
	proc = subprocess.run(zenity_cmd, capture_output=True)
	fn = proc.stdout.decode().rstrip()
	if fn == "":
		error_msg = "Opening canceled"
	else:
		model = wire3d.Group()
		load_file(fn)
		setup_view()

if len(sys.argv) > 1:
	try:
		load_file(sys.argv[1])
		setup_view()
	except OSError as e:
		print(e, file=sys.stderr)
		sys.exit(1)
	except SyntaxError as e:
		print(e, file=sys.stderr)
		sys.exit(2)
	except RuntimeError as e:
		print(e, file=sys.stderr)
		sys.exit(3)

while True:
	screen.fill(DARKSLATEGRAY)
	
	m2 = model.copy()
	m2.rotateZ(radians(angle_z)).rotateX(radians(angle_x))
	m2.render_through(camera, renderer)
	
	coord_msg = "x:{0} y:{1} z:{2}".format(camera.x, camera.z, camera.y)
	angle_msg = "u:{0} v:{1}".format(angle_z, angle_x)
	drawstr(coord_msg, FSIZE, FSIZE)
	drawstr(angle_msg, FSIZE, FSIZE * 2)
	drawstr(error_msg, FSIZE, SCREENH - FSIZE)
	
	pygame.display.flip()

	ev = pygame.event.wait()
	if ev.type == pygame.QUIT:
		break
	elif ev.type == pygame.VIDEORESIZE:
		SCREENW = ev.w
		SCREENH = ev.h
		CENTERX = SCREENW // 2
		CENTERY = SCREENH // 2
		FSIZE = SCREENH // 20
		
		screen = pygame.display.set_mode(
			(SCREENW, SCREENH), pygame.RESIZABLE)

		camera.f = min(SCREENW, SCREENH)
	elif ev.type == pygame.KEYUP:
		if ev.key in (pygame.K_ESCAPE, pygame.K_q):
			break
		elif ev.key in (pygame.K_EQUALS, pygame.K_KP_PLUS):
			if camera.z < bb_c.y:
				camera.z += zoom_step
		elif ev.key in (pygame.K_MINUS, pygame.K_KP_MINUS):
			camera.z -= zoom_step
		elif ev.key == pygame.K_LEFT:
			angle_z += ROTSTEP
			if angle_z > 180:
				angle_z -= 360
		elif ev.key == pygame.K_RIGHT:
			angle_z -= ROTSTEP
			if angle_z < -180:
				angle_z += 360
		elif ev.key == pygame.K_UP:
			if angle_x < 90:
				angle_x += ROTSTEP
		elif ev.key == pygame.K_DOWN:
			if angle_x > -90:
				angle_x -= ROTSTEP
		elif ev.key == pygame.K_r:
			angle_x = 0
			angle_z = 0
			camera.z = bb_c.y - max(size.x, size.z) * 1.5
			if camera.z >= m_min.y:
				camera.z = bb_c.y - size.y
		elif ev.key == pygame.K_o:
			if shutil.which("zenity") == None:
				error_msg = "Please install Zenity"
			else:
				open_with_zenity()
